package base;

import java.io.*;
import java.net.*;

import request.*;

import util.*;

import config.*;


///////////////////////////////////////////////////////////////////////////////////
//
///////////////////////////////////////////////////////////////////////////////////
public class ClientHandler extends Thread
{
	private Socket socket;
	private Server server;
	
	private static final int CYCLE_BUFFER_SIZE = 4096;


	
	///////////////////////////////////////////////////////////////////////////////////
	//
	///////////////////////////////////////////////////////////////////////////////////
	public ClientHandler(Server in_server, Socket in_socket)
	{
		server = in_server;
		socket = in_socket;	
	}
    
	///////////////////////////////////////////////////////////////////////////////////
	//
	///////////////////////////////////////////////////////////////////////////////////
	public void run() 
	{
		InputStream in = null;
		OutputStream out = null;
		
		byte[] data_response = null;
		byte[] data_request = null;

		try
    	{
			in = socket.getInputStream();
			out = socket.getOutputStream();

			data_request = getRequestFromClient(in,out);
						
			if (data_request == null)
			{
				//bypass invalid request
			}
			else if (data_request[0] == Request.ID_HANDLE_POLICY_FILE)
	    	{
				UtilLog.write("        INFO: sent policy file to: " + socket.getInetAddress());
	    	}
	    	else
	    	{	    	
	    		//decrypt from the flash client
				UtilCipher.decrypt(data_request);

	    		data_response = processRequest(data_request);
    			sendResponseToClient(out,data_response);
	    	}

    	}
		catch (Exception e)
		{
			UtilEmail.postMail(Config.email_for_logging, "error processing request from: " + socket.getInetAddress(),"",Config.email_for_logging);
			UtilLog.write("        ERROR: from " + socket.getInetAddress() + ":" + e.getMessage());
			e.printStackTrace();
    	}
		finally
		{
			try
			{
		    	if (in != null)
		    		in.close();
				
		    	if (out != null)
		    		out.close();
	    		
		    	if (socket != null)
		    		socket.close();	
			}
			catch(Exception ee)
			{
				UtilLog.write("        ERROR: unable to close handles for " + socket.getInetAddress());
				ee.printStackTrace();
			}
		}

		try
		{
			updateAudits(data_request);
		}
		catch(Exception e)
		{
			UtilLog.write("        ERROR: unable to write out audits for " + socket.getInetAddress());
		}

		server.removeClientHandler(this);
	}

	

	
	///////////////////////////////////////////////////////////////////////////////////
	//
	///////////////////////////////////////////////////////////////////////////////////
	private void sendPolicyFile(OutputStream out) throws Exception
	{
		 
		String str = "<?xml version=\"1.0\"?>" + 
					 "<!DOCTYPE cross-domain-policy SYSTEM \"http://www.macromedia.com/xml/dtds/cross-domain-policy.dtd\">" +
					 "<cross-domain-policy>" + 
					 	"<allow-access-from domain=\"*\" to-ports=\"" + Config.port + "\" />" + 
					 "</cross-domain-policy>\0";

		out.write(str.getBytes());

	}


	///////////////////////////////////////////////////////////////////////////////////
	//
	//
	///////////////////////////////////////////////////////////////////////////////////
	private byte[] getRequestFromClient(InputStream in, OutputStream out) throws Exception
	{
		
		int i;
		byte request_id;
		byte decrypted_request_id;

		try
		{
			ByteArrayOutputStream data = new ByteArrayOutputStream();

			//do a blocking read to get the 1st byte
			//which is the request id
			request_id = (byte)(in.read());

			//decrypt the request id from the flash client
			if (request_id == 60)
				decrypted_request_id = request_id;
			else
				decrypted_request_id = UtilCipher.decrypt(request_id);
			
			if (!Request.isValid(decrypted_request_id))
				return null;

			//sending policy file back here 
			//because if code does it in the policy file message
			//or after the next read, it won't work
			//not entirely sure why
			if (decrypted_request_id == Request.ID_HANDLE_POLICY_FILE)
			{
				sendPolicyFile(out);
				return null;
			}
			
			//cycle through data - via a temp_data buffer -
			//and push into the bytearrayoutputstream.
			//we are doing this because we don't know the
			//size of the data. the bytearrayoutputstream
			//will grow dynamically to handle it
			int val;

			//write out request id
			data.write(request_id);

			//size of cycle buffer can affect efficiency
			byte[] temp_data = new byte[CYCLE_BUFFER_SIZE];


			//write out rest of data
			val = in.read(temp_data,0,temp_data.length);

			while(val != -1)
			{				
				for(i=0;i<val;i++)
					data.write(temp_data[i]);

				if (in.available() < 1)
					break;
				
				val = in.read(temp_data,0,temp_data.length);
			}
			

			if (data.toByteArray().length < 1)
				throw new Exception("WARNING: data had a length of:" + data.toByteArray().length + " from client " + socket.getInetAddress());

			
			return data.toByteArray();
		}
		catch(Exception e)
		{
			throw new Exception("ERROR: unable to get request from client " + socket.getInetAddress());
		}
	}


	///////////////////////////////////////////////////////////////////////////////////
	//
	///////////////////////////////////////////////////////////////////////////////////
	private byte[] processRequest(byte[] data) throws Exception
	{
		Request req = Request.factory(data);

		String str = "processing " + req.toString() + " for " + socket.getInetAddress() + "\n";
		str = str.trim();
		str = "        " + str;
		UtilLog.write(str);
		
		byte[] data_response = req.process();
		
		//dsd make this into audits
		if (req instanceof RequestSendClientInfo)
			((RequestSendClientInfo)(req)).display(socket);

		return data_response;
	}
    

	///////////////////////////////////////////////////////////////////////////////////
	//
	///////////////////////////////////////////////////////////////////////////////////
	private void sendResponseToClient(OutputStream out, byte[] data) throws Exception
	{
		//encrypt to the flash client
		UtilCipher.encrypt(data);
		
		try
		{
			out.write(data);
		}
		catch(Exception e)
		{
			throw new Exception("ERROR: unable to send response to client " + socket.getInetAddress());
		}
	}


	
	
	///////////////////////////////////////////////////////////////////////////////////
	//
	///////////////////////////////////////////////////////////////////////////////////
	private void updateAudits(byte[] data) throws Exception
	{
		if (data == null)
			return;
	
		//dsd store these audits in the db
		if (data[0] == Request.ID_GET_APP_INFO)
		{
			UtilEmail.postMail(Config.email_for_logging, "user got app info: " + socket.getInetAddress(),"",Config.email_for_logging);
		}
	}
	
}
